/*
 *  Main header file for the ULTRA driver
 *  Copyright (c) 1994-98 by Jaroslav Kysela <perex@jcu.cz>
 */

#include "config.h"

#define VERSION \
"Version " ULTRA_VERSION

#define COPYRIGHT \
"Copyright (c) 1994-98 by Jaroslav Kysela <perex@jcu.cz>."

#define ULTRA_CARDS		8	/* number of supported ULTRA cards - don't change - minor numbers */

#include "config.h"
#ifndef ULTRACFG_MAJOR			/* standard configuration */
#define ULTRACFG_MAJOR		14
#endif

#if !defined( ULTRACFG_DEBUG )
#undef ULTRACFG_DEBUG_MEMORY
#endif

/*
 *  ==========================================================================
 */

#ifdef LINUX

#define __KERNEL__
#define MODULE

#define LinuxVersionCode( v, p, s ) (((v)<<16)|((p)<<8)|(s))

#include <linux/config.h>
#include <linux/version.h>
#if defined( CONFIG_MODVERSIONS )
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#ifndef ULTRA_MAIN_OBJECT_FILE
#define __NO_VERSION__
#endif
#include <linux/module.h>

#if LinuxVersionCode( 2, 0, 0 ) > LINUX_VERSION_CODE
#error "This driver is designed only for Linux 2.0.0 and highter."
#endif
#if LinuxVersionCode( 2, 1, 0 ) <= LINUX_VERSION_CODE
#define LINUX_2_1
#endif
/* notice: I don't know which kernel have f_dentry */
#if LinuxVersionCode( 2, 1, 45 ) <= LINUX_VERSION_CODE
#define LINUX_2_1_X
#endif

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/malloc.h>

#include <linux/ioport.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/dma.h>
#include <asm/segment.h>
#ifdef LINUX_2_1
#include <asm/uaccess.h>
#endif
#include <asm/system.h>
#include <asm/string.h>

#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#ifdef LINUX_2_1
#include <linux/vmalloc.h>
#endif

#undef ULTRA_POLL
#if LinuxVersionCode( 2, 1, 72 ) <= LINUX_VERSION_CODE
#define ULTRA_POLL
#endif
#ifdef __MAX_POLL_TABLE_ENTRIES		/* new kernel with poll() */
#define ULTRA_POLL
#endif
#ifdef ULTRA_POLL
#include <linux/poll.h>
#endif

#if LinuxVersionCode( 2, 1, 35 ) >= LINUX_VERSION_CODE
#define in_interrupt() intr_count
#endif

#include "ultra.h"

#define PRINTK printk

#define OUTB( value, reg ) outb( value, reg )
#define OUTW( value, reg ) outw( value, reg )
#define INB( reg ) inb( reg )
#define INW( reg ) inw( reg )
#define OUTSB( reg, ptr, count ) outsb( reg, ptr, count )
#define OUTSW( reg, ptr, count ) outsw( reg, ptr, count )
#define INSB( reg, ptr, count ) insb( reg, ptr, count )
#define INSW( reg, ptr, count ) insw( reg, ptr, count )
static inline void CLI( unsigned long *flags ) { save_flags( *flags ); cli(); }
static inline void STI( unsigned long *flags ) { restore_flags( *flags ); }

#define MB() 			/* 386+ architecture doesn't need io sync */

#define SLEEP_DEFINE( ident ) \
  struct wait_queue *sleeper_##ident; \
  unsigned long end_jiffies_##ident; \
  unsigned short sleeper_lock_##ident;
#define SLEEP_PREPARE( card, ident ) \
  card -> sleeper_##ident = NULL;
#define SLEEP( card, ident, time ) \
  card -> end_jiffies_##ident = current -> timeout = jiffies + (time); \
  interruptible_sleep_on( &(card) -> sleeper_##ident );
#define SLEEPU( card, ident, time ) \
  card -> end_jiffies_##ident = current -> timeout = jiffies + (time); \
  sleep_on( &(card) -> sleeper_##ident );
#ifdef ULTRA_POLL
#define SLEEP_POLL( card, ident, table ) \
  poll_wait( &(card) -> sleeper_##ident, table )
#else
#define SLEEPS( card, ident, table ) \
  select_wait( &(card) -> sleeper_##ident, table );
#endif
#if LinuxVersionCode( 2, 1, 71 ) <= LINUX_VERSION_CODE
  #define TABORT( card, ident ) \
    ( signal_pending( current ) )
#else
  #define TABORT( card, ident ) \
    ( current -> signal & ~current -> blocked )
#endif
#define WAKEUP( card, ident ) wake_up( &(card) -> sleeper_##ident );
#define GETLOCK( card, ident ) (card) -> sleeper_lock_##ident
#define TIMEOUT( card, ident ) ( (card) -> end_jiffies_##ident < jiffies )
#define TIMEOUT_VALUE( card, ident ) ( (card) -> end_jiffies_##ident - jiffies )
#define MUTEX_DEFINE( ident ) \
  struct semaphore mutex_##ident;
#define MUTEX_PREPARE( card, ident ) \
  (card) -> mutex_##ident = MUTEX;
#define MUTEX_DOWN( card, ident ) down( &(card) -> mutex_##ident )
#define MUTEX_UP( card, ident ) up( &(card) -> mutex_##ident )
#define MUTEX_DEFINE_STATIC( ident ) \
  static struct semaphore local_mutex_##ident = MUTEX;
#define MUTEX_DOWN_STATIC( ident ) down( &local_mutex_##ident )
#define MUTEX_UP_STATIC( ident ) up( &local_mutex_##ident )

#define VERIFY_AREA verify_area

#define MEMCPY memcpy
#define MEMSET memset
#define MEMMOVE memmove

#ifdef LINUX_2_1
#define MEMCPY_FROMFS copy_from_user
#define MEMCPY_TOFS copy_to_user
#else
#define MEMCPY_FROMFS memcpy_fromfs
#define MEMCPY_TOFS memcpy_tofs
#endif

#define VMALLOC vmalloc
#define VFREE vfree

#ifdef LINUX_2_1

static inline void put_fs_byte( unsigned char val, char *ptr ) { put_user( val, ptr ); }
static inline void put_fs_word( unsigned short val, short *ptr ) { put_user( val, ptr ); }
static inline unsigned char get_fs_byte( unsigned char *ptr ) { unsigned char val; get_user( val, ptr ); return val; }
static inline unsigned short get_fs_word( unsigned short *ptr ) { unsigned short val; get_user( val, ptr ); return val; }

#else

#define get_fs_int		get_fs_long
#define put_fs_int		put_fs_long

#endif

#define IOCTL_IN( arg )		ultra_ioctl_in( (long *)arg )
#define IOCTL_OUT( arg, val )	ultra_ioctl_out( (long *)arg, val )

#define SA_ULTRA_FLAGS		(SA_INTERRUPT)
#define ultra_request_irq	request_irq
#define ultra_free_irq		free_irq

#else

#error Unknown system!!!

#endif

/*
 *  ==========================================================================
 */

#include "misc.h"

#define SNDSTAT_STRING \
"GUS driver - compatibility text for gmod : 3.0\n"

/*
 *  MINOR numbers
 */

#define ULTRA_MINOR_MIXER	0	/* /dev/mixer - OSS 3.XX compatible */
#define ULTRA_MINOR_SEQUENCER	1	/* /dev/sequencer - OSS 3.XX compatible */
#define	ULTRA_MINOR_MIDI	2	/* /dev/midi - native midi interface - OSS 3.XX compatible - UART */
#define ULTRA_MINOR_PCM_8	3	/* /dev/dsp - 8bit PCM - OSS 3.XX compatible */
#define ULTRA_MINOR_AUDIO	4	/* /dev/audio - SunSparc compatible */
#define ULTRA_MINOR_PCM_16	5	/* /dev/dsp16 - 16bit PCM - OSS 3.XX compatible */
#define ULTRA_MINOR_SNDSTAT	6	/* /dev/sndstat - for compatibility with OSS */
#define ULTRA_MINOR_RESERVED7	7	/* reserved for future use */
#define ULTRA_MINOR_MUSIC	8	/* /dev/music - OSS 3.XX compatible */
#define ULTRA_MINOR_DMMIDI	9	/* /dev/dmmidi0 - this device can have another minor # with OSS */
#define ULTRA_MINOR_DMFM	10	/* /dev/dmfm0 - this device can have another minor # with OSS */
#define ULTRA_MINOR_MIXER1	11	/* alternate mixer */
#define ULTRA_MINOR_PCM1	12	/* alternate PCM (GF-A-1) */
#define ULTRA_MINOR_MIDI1	13	/* alternate midi - SYNTH */
#define ULTRA_MINOR_DMMIDI1	14	/* alternate dmmidi - SYNTH */

#define ULTRA_MINOR_OSS_MASK	0x000f

#define ULTRA_MINOR_UBEGIN	128

#define ULTRA_MINOR_USYNTH	0	/* /dev/ultra* - ultra synth */
#define ULTRA_MINOR_UCONTROL	1	/* /dev/ultractl* - ultra control */
#define ULTRA_MINOR_USERVER	2	/* /dev/ultrasrv* - instrument server interface */
#define ULTRA_MINOR_UMIDI	3	/* /dev/ultramidi* - midi */
#define ULTRA_MINOR_UFM		4	/* /dev/ultrafm* - FM/OPL3 */
#define ULTRA_MINOR_UMIXER0	6	/* /dev/ultramixer*0 - mixer 0 */
#define ULTRA_MINOR_UMIXER1	7	/* /dev/ultramixer*1 - mixer 1 */
#define ULTRA_MINOR_UPCM0	8	/* /dev/ultrapcm*0 - pcm 0 */
#define ULTRA_MINOR_UPCM1	9	/* /dev/ultrapcm*1 - pcm 1 */
#define ULTRA_MINOR_UMIDI_RAW0	10	/* /dev/ultramidi*0 - raw midi 0 */
#define ULTRA_MINOR_UMIDI_RAW1	11	/* /dev/ultramidi*1 - raw midi 1 */
#define ULTRA_MINOR_UDETECT	14	/* /dev/ultradetect - detection info */
#define ULTRA_MINOR_UINFO	15	/* /dev/ultrainfo - show info */

#define ULTRA_MINOR_UDEVMASK	0x0f

/* auto value */

#define ULTRA_AUTO		0x100	/* seems to be reasonable */
					/* ISA bus adapters doesn't use this number for port */

/* sleep & wakeup state */

#define WK_NONE			0x00
#define WK_LOCK			0x01		/* lock this DMA channel */
#define WK_WAKEUP		0x02		/* DMA transfer done */
#define WK_SLEEP		0x04		/* DMA sleep */

/* kernel / user space */

#define SP_KERNEL		0x00
#define SP_USER			0x01
#define SP_KERNEL_WAVE_USER	0x02		/* special for OSS Sequencer */

/* Hardware resources allocation */

#define ULTRA_DMAS	4	/* max DMA # for soundcard */
#define ULTRA_IRQS	4	/* max IRQ # for soundcard */
#define ULTRA_PORTS	8	/* max PORT regions for soundcard */

struct ULTRA_STRU_DMA {
  unsigned short static_alloc: 1, /* pre-allocated */
  		 mmaped: 1;	  /* mmaped area */
  short used;			/* dma channel is used (count) */
  short dma;			/* DMA number */
  char *name;			/* pointer to name */
  unsigned char *buf;		/* pointer to DMA buffer */
  long size;			/* real size of DMA buffer */
  long rsize;			/* requested size of DMA buffer */
  long usize;			/* used size of DMA buffer */
  long ursize;			/* used requested size of DMA buffer */
  unsigned short lock;		/* lock flags for this DMA channel */
  char *owner;			/* owner of this DMA channel */
  char *old_owner;		/* old owner of this DMA channel */
};

struct ULTRA_STRU_IRQ {
  unsigned short number;
  char *name;
};

struct ULTRA_STRU_CARD_PORT_REGION {
  unsigned short port;
  unsigned short size;
  char *name;
};

typedef void (ultra_irq_handler_t)( int irq, void *dev_id, struct pt_regs *regs );

/* main structure for GUS card */

struct ULTRA_STRU_CARD {
  int number;			/* number of soundcard (index to ultra_cards) */

  unsigned int version;		/* version (number ID) of soundcard */

  char id[ 8 ];			/* id string of this card */
  char name[ 64 ];		/* name of this soundcard */
  char info1[ 80 ];		/* info1 about this soundcard */
  char info2[ 80 ];		/* info2 about this soundcard */

  struct ULTRA_STRU_DMA dmas[ ULTRA_DMAS ];
  unsigned int dmas_map;	/* bitmask for allocated DMA channels */
  
  struct ULTRA_STRU_IRQ irqs[ ULTRA_IRQS ];
  unsigned int irqs_map;	/* bitmask for allocated IRQ channels */

  struct ULTRA_STRU_CARD_PORT_REGION ports[ ULTRA_PORTS ];
  unsigned int ports_map;	/* bitmask for allocated port numbers */

  void (*use_inc)( void );	/* increment use count */
  void (*use_dec)( void );	/* decrement use count */
  
  void *private;		/* private structure for soundcard */
  void (*private_free)( void *private ); /* callback for freeing of private structure */
};

typedef struct ULTRA_STRU_CARD ultra_card_t;

/* buffer for informations */

struct ultra_info_buffer {
  char *buffer;
  char *curr;
  unsigned long size;
  int stop;
};

typedef struct ultra_info_buffer ultra_info_buffer_t;

/* device.c */

typedef int (ultra_unregister_t)( void );

typedef long long (ultra_lseek_t)( struct file *file, long long offset, int orig );
typedef long (ultra_read_t)( struct file *file, char *buf, long count );
typedef long (ultra_write_t)( struct file *file, const char *buf, long count );
typedef int (ultra_open_t)( unsigned short minor, struct file *file );
typedef int (ultra_release_t)( unsigned short minor, struct file *file );
#ifdef ULTRA_POLL
typedef unsigned int (ultra_poll_t)( struct file *file, poll_table *wait );
#else
typedef int (ultra_select_t)( struct file *file, int sel_type, select_table *wait );
#endif
typedef int (ultra_ioctl_t)( struct file *file, unsigned int cmd, unsigned long arg );
typedef int (ultra_mmap_t)( struct inode *inode, struct file *file, struct vm_area_struct *vma );

struct ultra_stru_register_minor {
  char *comment;		/* for /dev/ultrainfo */

  int verify_card;		/* verify access to soundcard */

  int ultra_minors[ 16 ];
  int oss_minors[ 16 ];

  ultra_unregister_t *ultra_unregister;

  ultra_lseek_t *ultra_lseek;
  ultra_read_t *ultra_read;
  ultra_write_t *ultra_write;
  ultra_open_t *ultra_open;
  ultra_release_t *ultra_release;
#ifdef ULTRA_POLL
  ultra_poll_t *ultra_poll;
#else
  ultra_select_t *ultra_select;
#endif
  ultra_ioctl_t *ultra_ioctl;
  ultra_mmap_t *ultra_mmap;
};

extern int ultra_register_minor( struct ultra_stru_register_minor *reg );
extern int ultra_unregister_minor( struct ultra_stru_register_minor *reg );

extern void ultra_minor_info( ultra_info_buffer_t *buffer );

extern int ultra_ioctl_in( long *addr );
extern int ultra_ioctl_out( long *addr, int value );

/* memory.c */

extern void ultra_malloc_init( void );
extern void ultra_malloc_done( void );
extern void *ultra_malloc( unsigned long size );
extern void ultra_free( void *obj, unsigned long size );
extern char *ultra_malloc_strdup( char *string, int space );
extern void ultra_free_str( char *string );
extern char *ultra_malloc_pages( unsigned long size, int *pg, int dma );
extern void ultra_free_pages( char *ptr, unsigned long size );
extern int ultra_dma_malloc( ultra_card_t *card, int idx, char *owner, int lock );
extern void ultra_dma_free( ultra_card_t *card, int idx, int lock );
extern void ultra_memory_info( ultra_info_buffer_t *buffer );
#ifdef ULTRACFG_DEBUG_MEMORY
extern void ultra_memory_debug( ultra_info_buffer_t *buffer );
#endif

/* init.c */

extern int ultra_cards_count;
extern unsigned int ultra_cards_bitmap;
extern ultra_card_t *ultra_cards[ ULTRA_CARDS ];
extern ultra_card_t *ultra_cards_irq_index[ 16 ];

extern void ultra_driver_init( void );

extern ultra_card_t *ultra_card_new( char *id, char *name, void (*use_inc)( void ), void (*use_dec)( void ) );
extern int ultra_card_free( ultra_card_t *card );
extern int ultra_card_register( ultra_card_t *card );
extern int ultra_card_unregister( ultra_card_t *card );
extern char *ultra_get_id( ultra_card_t *card, char *id );
extern void ultra_info_card( ultra_info_buffer_t *buffer );

extern void ultra_delay( int loops );

extern int ultra_register_ioport( ultra_card_t *card, int port, int size, char *name );
extern int ultra_unregister_ioports( ultra_card_t *card );

extern int ultra_register_dma_channel( ultra_card_t *card, int idx, char *name, int number, int rsize, int *possible_numbers );
extern int ultra_unregister_dma_channels( ultra_card_t *card );

extern int ultra_register_interrupt( ultra_card_t *card, int number, ultra_irq_handler_t *handler, char *name, int *possible_number );
extern int ultra_unregister_interrupts( ultra_card_t *card );

/* info.c */

extern int ultra_iprintf( ultra_info_buffer_t *buffer, char *fmt, ... );
extern int ultra_info_register( void );

/* control.c */

extern int ultra_control_register( void );

/* --- */

#ifdef ULTRACFG_DEBUG
#define PRINTD( args... ) PRINTK( ##args )
#else
#define PRINTD( args... ) /* nothing */
#endif
