/*
 *  Global structures used for GUS part of ULTRA driver
 */

#include "ultramemory.h"

#define ULTRA_GUS_DMA_GPLAY	0x0000
#define ULTRA_GUS_DMA_GRECORD	0x0001
#define ULTRA_GUS_DMA_CPLAY	0x0001
#define ULTRA_GUS_DMA_CRECORD	0x0000
#define ULTRA_GUS_DMA_DB	0x0002
#define ULTRA_GUS_DMA_ESS	0x0001

#define ULTRA_GUS_IRQ_GF1	0x0000
#define ULTRA_GUS_IRQ_DB	0x0001		/* daughter board */
#define ULTRA_GUS_IRQ_ESS	0x0001


/* IO ports */

#define ULTRA_GUSP( gus, x )		(gus -> gf1.port + ULTRA_g_u_s_##x)

#define ULTRA_g_u_s_MIDICTRL		(0x320-0x220)
#define ULTRA_g_u_s_MIDISTAT		(0x320-0x220)
#define ULTRA_g_u_s_MIDIDATA		(0x321-0x220)

#define ULTRA_g_u_s_GF1PAGE		(0x322-0x220)
#define ULTRA_g_u_s_GF1REGSEL		(0x323-0x220)
#define ULTRA_g_u_s_GF1DATALOW		(0x324-0x220)
#define ULTRA_g_u_s_GF1DATAHIGH		(0x325-0x220)
#define ULTRA_g_u_s_IRQSTAT		(0x226-0x220)
#define ULTRA_g_u_s_TIMERCNTRL		(0x228-0x220)
#define ULTRA_g_u_s_TIMERDATA		(0x229-0x220)
#define ULTRA_g_u_s_DRAM		(0x327-0x220)
#define ULTRA_g_u_s_MIXCNTRLREG		(0x220-0x220)
#define ULTRA_g_u_s_IRQDMACNTRLREG	(0x22b-0x220)
#define ULTRA_g_u_s_REGCNTRLS		(0x22f-0x220)
#define ULTRA_g_u_s_BOARDVERSION	(0x726-0x220)
#define ULTRA_g_u_s_MIXCNTRLPORT	(0x726-0x220)
#define ULTRA_g_u_s_IVER		(0x325-0x220)
#define ULTRA_g_u_s_MIXDATAPORT		(0x326-0x220)
#define ULTRA_g_u_s_MAXCNTRLPORT	(0x326-0x220)

/* GF1 registers */

/* global registers */
#define ULTRA_GF1_GB_ACTIVE_VOICES		0x0e
#define ULTRA_GF1_GB_VOICES_IRQ			0x0f
#define ULTRA_GF1_GB_GLOBAL_MODE		0x19
#define ULTRA_GF1_GW_LFO_BASE			0x1a
#define ULTRA_GF1_GB_VOICES_IRQ_READ		0x1f
#define ULTRA_GF1_GB_DRAM_DMA_CONTROL		0x41
#define ULTRA_GF1_GW_DRAM_DMA_LOW		0x42
#define ULTRA_GF1_GW_DRAM_IO_LOW		0x43
#define ULTRA_GF1_GB_DRAM_IO_HIGH		0x44
#define ULTRA_GF1_GB_SOUND_BLASTER_CONTROL	0x45
#define ULTRA_GF1_GB_ADLIB_TIMER_1		0x46
#define ULTRA_GF1_GB_ADLIB_TIMER_2		0x47
#define ULTRA_GF1_GB_REC_DMA_CONTROL		0x49
#define ULTRA_GF1_GB_JOYSTICK_DAC_LEVEL		0x4b
#define ULTRA_GF1_GB_RESET			0x4c
#define ULTRA_GF1_GB_DRAM_DMA_HIGH		0x50
#define ULTRA_GF1_GW_DRAM_IO16			0x51
#define ULTRA_GF1_GW_MEMORY_CONFIG		0x52
#define ULTRA_GF1_GB_MEMORY_CONTROL		0x53
#define ULTRA_GF1_GW_FIFO_RECORD_BASE_ADDR	0x54
#define ULTRA_GF1_GW_FIFO_PLAY_BASE_ADDR	0x55
#define ULTRA_GF1_GW_FIFO_SIZE			0x56
#define ULTRA_GF1_GW_INTERLEAVE			0x57
#define ULTRA_GF1_GB_COMPATIBILITY		0x59
#define ULTRA_GF1_GB_DECODE_CONTROL		0x5a
#define ULTRA_GF1_GB_VERSION_NUMBER		0x5b
#define ULTRA_GF1_GB_MPU401_CONTROL_A		0x5c
#define ULTRA_GF1_GB_MPU401_CONTROL_B		0x5d
#define ULTRA_GF1_GB_EMULATION_IRQ		0x60
/* voice specific registers */
#define ULTRA_GF1_VB_ADDRESS_CONTROL		0x00
#define ULTRA_GF1_VW_FREQUENCY			0x01
#define ULTRA_GF1_VW_START_HIGH			0x02
#define ULTRA_GF1_VW_START_LOW			0x03
#define ULTRA_GF1_VA_START			GF1_VW_START_HIGH
#define ULTRA_GF1_VW_END_HIGH			0x04
#define ULTRA_GF1_VW_END_LOW			0x05
#define ULTRA_GF1_VA_END			GF1_VW_END_HIGH
#define ULTRA_GF1_VB_VOLUME_RATE		0x06
#define ULTRA_GF1_VB_VOLUME_START		0x07
#define ULTRA_GF1_VB_VOLUME_END			0x08
#define ULTRA_GF1_VW_VOLUME			0x09
#define ULTRA_GF1_VW_CURRENT_HIGH		0x0a
#define ULTRA_GF1_VW_CURRENT_LOW		0x0b
#define ULTRA_GF1_VA_CURRENT			GF1_VW_CURRENT_HIGH
#define ULTRA_GF1_VB_PAN			0x0c
#define ULTRA_GF1_VW_OFFSET_RIGHT		0x0c
#define ULTRA_GF1_VB_VOLUME_CONTROL		0x0d
#define ULTRA_GF1_VB_UPPER_ADDRESS		0x10
#define ULTRA_GF1_VW_EFFECT_HIGH		0x11
#define ULTRA_GF1_VW_EFFECT_LOW			0x12
#define ULTRA_GF1_VA_EFFECT			GF1_VW_EFFECT_HIGH
#define ULTRA_GF1_VW_OFFSET_LEFT		0x13
#define ULTRA_GF1_VB_ACCUMULATOR		0x14
#define ULTRA_GF1_VB_MODE			0x15
#define ULTRA_GF1_VW_EFFECT_VOLUME		0x16
#define ULTRA_GF1_VB_FREQUENCY_LFO		0x17
#define ULTRA_GF1_VB_VOLUME_LFO			0x18
#define ULTRA_GF1_VW_OFFSET_RIGHT_FINAL		0x1b
#define ULTRA_GF1_VW_OFFSET_LEFT_FINAL		0x1c
#define ULTRA_GF1_VW_EFFECT_VOLUME_FINAL	0x1d

/* ICS registers */

#define ICS_MIC_DEV		0
#define ICS_LINE_DEV		1
#define ICS_CD_DEV		2
#define ICS_GF1_DEV		3
#define ICS_NONE_DEV		4
#define ICS_MASTER_DEV		5

/* LFO */

#define ULTRA_LFO_TREMOLO		0
#define ULTRA_LFO_VIBRATO		1

/* misc */

#define ULTRA_GUS_DMA_DOWNLOAD_SLEEP_TIME	(HZ*2)
#define ULTRA_GUS_BAD_DMA_DOWNLOAD_SLEEP_TIME	15

/* ramp ranges */

#define ULTRA_ATTEN( x )	(ultra_atten_table[ x ])
#define ULTRA_VOLUME( x )	(ultra_atten_table[ x ] << 1)
#define ULTRA_MIN_VOLUME	1800
#define ULTRA_MAX_VOLUME	4095
#define ULTRA_MIN_OFFSET	(MIN_VOLUME>>4)
#define ULTRA_MAX_OFFSET	255
#define ULTRA_MAX_TDEPTH	90

/* indexs for voice_ranges array */

#define ULTRA_GF1_VOICE_RANGES        3
#define ULTRA_GF1_VOICE_RANGE_SYNTH   0
#define ULTRA_GF1_VOICE_RANGE_PCM     1
#define ULTRA_GF1_VOICE_RANGE_EFFECT  2

/* constants for interrupt handlers */

#define ULTRA_GF1_HANDLER_MIDI_OUT            0x00010000
#define ULTRA_GF1_HANDLER_MIDI_IN             0x00020000
#define ULTRA_GF1_HANDLER_TIMER1              0x00040000
#define ULTRA_GF1_HANDLER_TIMER2              0x00080000
#define ULTRA_GF1_HANDLER_RANGE               0x00100000
#define ULTRA_GF1_HANDLER_SYNTH_DMA_WRITE     0x00200000
#define ULTRA_GF1_HANDLER_PCM_DMA_WRITE       0x00400000
#define ULTRA_GF1_HANDLER_PCM_DMA_READ        0x00800000
#define ULTRA_GF1_HANDLER_ALL                 (0xffff0000&~ULTRA_GF1_HANDLER_RANGE)

/* --- */

struct ULTRA_STRU_GUS_CARD;
typedef struct ULTRA_STRU_GUS_CARD ultra_gus_card_t;

/* GF1 specific structure */

struct ULTRA_STRU_GF1 {

  unsigned int enh_mode: 1,		/* enhanced mode (GFA1) */
               pcm_memory: 1,		/* reserve memory for PCM */ 
  	       dma_download: 1,		/* use DMA download to GUS's DRAM */
  	       hw_lfo: 1,		/* use hardware LFO */
  	       sw_lfo: 1;		/* use software LFO */
  
  unsigned int port;		/* port of GF1 chip */
  unsigned int irq;		/* IRQ number of GF1 chip */
  unsigned int memory;		/* GUS's DRAM size in bytes */
  unsigned int rom_memory;	/* GUS's ROM size in bytes */
  unsigned int rom_present;	/* bitmask */
  unsigned int rom_banks;	/* GUS's ROM banks */

  struct ULTRA_STRU_BANK_INFO banks_8 [ 4 ];
  struct ULTRA_STRU_BANK_INFO banks_16[ 4 ];    

  /* registers */
  unsigned short reg_page;
  unsigned short reg_regsel;
  unsigned short reg_data8;
  unsigned short reg_data16;
  unsigned short reg_irqstat;
  unsigned short reg_dram;
  unsigned short reg_timerctrl;
  unsigned short reg_timerdata;
  /* --------- */

  unsigned char active_voice;	/* selected voice (GF1PAGE register) */
  unsigned char active_voices;	/* all active voices */

  unsigned int default_voice_address;

  unsigned short playback_freq;	/* GF1 playback (mixing) frequency */
  unsigned short mode;		/* see to GF1_MODE_XXXX */

  unsigned char *lfos;

  /* special interrupt handlers / callbacks */
  
  struct GF1_STRU_VOICE_RANGE {
    unsigned short rvoices;     /* requested voices */
    unsigned short voices;      /* total voices for this range */
    unsigned short min, max;    /* minimal & maximal voice */
#ifdef ULTRACFG_INTERRUPTS_PROFILE
    unsigned int interrupt_stat_wave;
    unsigned int interrupt_stat_volume;
#endif
    void (*interrupt_handler_wave)( ultra_gus_card_t *gus, int voice );
    void (*interrupt_handler_volume)( ultra_gus_card_t *gus, int voice );
    void (*voices_change_start)( ultra_gus_card_t *gus );
    void (*voices_change_stop)( ultra_gus_card_t *gus );
    void (*volume_change)( ultra_gus_card_t *gus );
  } voice_ranges[ ULTRA_GF1_VOICE_RANGES ];

  /* interrupt handlers */
  
  void (*interrupt_handler_midi_out)( ultra_card_t *card );
  void (*interrupt_handler_midi_in)( ultra_card_t *card );
  void (*interrupt_handler_timer1)( ultra_card_t *card );
  void (*interrupt_handler_timer2)( ultra_card_t *card );
  void (*interrupt_handler_synth_dma_write)( ultra_card_t *card );
  void (*interrupt_handler_pcm_dma_write)( ultra_card_t *card );
  void (*interrupt_handler_pcm_dma_read)( ultra_card_t *card );
};

/* main structure for GUS card */

struct ULTRA_STRU_GUS_CARD {
  ultra_card_t *card;

  unsigned int equal_irq:1,	/* GF1 and CODEC shares IRQ (GUS MAX only) */
  	       equal_dma:1,	/* if dma channels are equal (not valid for daughter board) */
  	       ics_flag:1,	/* have we ICS mixer chip */
  	       ics_flipped:1,	/* ICS mixer have flipped some channels? */
  	       codec_flag:1,	/* have we CODEC chip? */
  	       max_flag:1,	/* have we GUS MAX card? */
  	       max_ctrl_flag:1,	/* have we original GUS MAX card? */
  	       daughter_flag:1, /* have we daughter board? */
  	       interwave:1,	/* hey - we have InterWave card */ 
  	       ess_flag:1;	/* ESS chip detected */
  unsigned short max_cntrl_val;	/* GUS MAX control value */
  unsigned short mix_cntrl_reg;	/* mixer control register */
  unsigned short joystick_dac;	/* joystick DAC level */

  struct ULTRA_STRU_GF1 gf1;	/* gf1 specific variables */
};

/* I/O functions for GF1/InterWave chip */

static inline void ultra_gf1_select_voice( ultra_gus_card_t *gus, int voice )
{
  unsigned long flags;
  
  CLI( &flags );
  if ( voice != gus -> gf1.active_voice )
    {
      gus -> gf1.active_voice = voice;
      OUTB( voice, ULTRA_GUSP( gus, GF1PAGE ) );
    }
  STI( &flags );
}

extern void ultra_gf1_delay( ultra_gus_card_t *gus );

extern void ultra_gf1_ctrl_stop( ultra_gus_card_t *gus, unsigned char reg );

extern void ultra_gf1_write8( ultra_gus_card_t *gus, unsigned char reg, unsigned char data );
extern unsigned char ultra_gf1_look8( ultra_gus_card_t *gus, unsigned char reg );
extern inline unsigned char ultra_gf1_read8( ultra_gus_card_t *gus, unsigned char reg ) { return ultra_gf1_look8( gus, reg | 0x80 ); }
extern void ultra_gf1_write16( ultra_gus_card_t *gus, unsigned char reg, unsigned int data );
extern unsigned short ultra_gf1_look16( ultra_gus_card_t *gus, unsigned char reg );
extern inline unsigned short ultra_gf1_read16( ultra_gus_card_t *gus, unsigned char reg ) { return ultra_gf1_look16( gus, reg | 0x80 ); }
extern void ultra_gf1_adlib_write( ultra_gus_card_t *gus, unsigned char reg, unsigned char data );
extern void ultra_gf1_dram_addr( ultra_gus_card_t *gus, unsigned int addr );
extern void ultra_gf1_poke( ultra_gus_card_t *gus, unsigned int addr, unsigned char data );
extern unsigned char ultra_gf1_peek( ultra_gus_card_t *gus, unsigned int addr );
extern void ultra_gf1_pokew( ultra_gus_card_t *gus, unsigned int addr, unsigned short data );
extern unsigned short ultra_gf1_peekw( ultra_gus_card_t *gus, unsigned int addr );
extern void ultra_gf1_dram_setmem( ultra_gus_card_t *gus, unsigned int addr, unsigned short value, unsigned int count );
extern void ultra_gf1_write_addr( ultra_gus_card_t *gus, unsigned char reg, unsigned int addr, short w_16bit );
extern unsigned int ultra_gf1_read_addr( ultra_gus_card_t *gus, unsigned char reg, short w_16bit );
extern void ultra_gf1_i_ctrl_stop( ultra_gus_card_t *gus, unsigned char reg );
extern void ultra_gf1_i_write8( ultra_gus_card_t *gus, unsigned char reg, unsigned char data );
extern unsigned char ultra_gf1_i_look8( ultra_gus_card_t *gus, unsigned char reg );
extern void ultra_gf1_i_write16( ultra_gus_card_t *gus, unsigned char reg, unsigned int data );
extern inline unsigned char ultra_gf1_i_read8( ultra_gus_card_t *gus, unsigned char reg ) { return ultra_gf1_i_look8( gus, reg | 0x80 ); }
extern unsigned short ultra_gf1_i_look16( ultra_gus_card_t *gus, unsigned char reg );
extern inline unsigned short ultra_gf1_i_read16( ultra_gus_card_t *gus, unsigned char reg ) { return ultra_gf1_i_look16( gus, reg | 0x80 ); }
extern void ultra_gf1_i_adlib_write( ultra_gus_card_t *gus, unsigned char reg, unsigned char data );
extern void ultra_gf1_i_write_addr( ultra_gus_card_t *gus, unsigned char reg, unsigned int addr, short w_16bit );
extern unsigned int ultra_gf1_i_read_addr( ultra_gus_card_t *gus, unsigned char reg, short w_16bit );

extern void ultra_gf1_reselect_active_voices( ultra_gus_card_t *gus );

#ifdef ULTRACFG_DEBUG
extern void ultra_print_voice_registers( ultra_gus_card_t *gus );
extern void ultra_print_global_registers( ultra_gus_card_t *gus );
extern void ultra_print_setup_registers( ultra_gus_card_t *gus );
extern void ultra_peek_print_block( ultra_gus_card_t *gus, unsigned int addr, int count, int w_16bit );
#endif

ultra_gus_card_t *ultra_gus_new_card( ultra_card_t *card );
int ultra_gus_set_port( ultra_gus_card_t *card, unsigned short port );
int ultra_gus_detect_memory( ultra_gus_card_t *gus );
int ultra_gus_init_dma_irq( ultra_gus_card_t *gus, int latches );

void ultra_gus_interrupt( ultra_gus_card_t *gus, unsigned char status );

#if 0
extern void ultra_engine_instrument_register(
		unsigned short mode,
		struct ULTRA_STRU_INSTRUMENT_VOICE_COMMANDS *voice_cmds,
		struct ULTRA_STRU_INSTRUMENT_NOTE_COMMANDS *note_cmds,
		struct ULTRA_STRU_INSTRUMENT_CHANNEL_COMMANDS *channel_cmds );
extern int ultra_engine_instrument_register_ask( unsigned short mode );
#endif
