/*
 *  Interface file between ultra driver & user space
 *  Copyright (c) 1994-98 by Jaroslav Kysela <perex@jcu.cz>
 */

#ifndef __ULTRA_H
#define __ULTRA_H

#if defined( LINUX ) || defined( __LINUX__ ) || defined( __linux__ )
#include <asm/byteorder.h>
#include <linux/ioctl.h>
#ifdef __LITTLE_ENDIAN
#define ULTRA_LITTLE_ENDIAN
#endif
#endif

/*
 *  protocol version
 */

#define ULTRA_PROTOCOL_VERSION( major, minor, subminor ) ((major<<16)|(minor<<8)|subminor)

/*
 *  number of supported soundcards in one machine
 */

#define ULTRA_CARDS			8

/*
 *  types of soundcards
 */

#define ULTRA_CARD_TYPE_GUS_CLASSIC	0x00000024	/* revision 2.4 */
#define ULTRA_CARD_TYPE_GUS_CLASSIC1	0x00000034	/* revision 3.4? */
#define ULTRA_CARD_TYPE_GUS_CLASSIC_ICS 0x00000037	/* revision 3.7 (ICS mixer) */
#define ULTRA_CARD_TYPE_GUS_EXTREME	0x00000050	/* GUS Extreme */
#define ULTRA_CARD_TYPE_GUS_ACE		0x00000090	/* GUS ACE */
#define ULTRA_CARD_TYPE_GUS_MAX		0x000000a0	/* GUS MAX - revision 10 */
#define ULTRA_CARD_TYPE_GUS_MAX1	0x000000a1	/* GUS MAX - revision 11 */
#define ULTRA_CARD_TYPE_GUS_PNP		0x00000100	/* GUS Plug & Play with InterWave chip */
#define ULTRA_CARD_TYPE_IW_DYNASONIC	0x00000101	/* DynaSonic 3-D with InterWave chip */

/*
 *  MIDI controllers
 */
						/* emulation */
#define ULTRA_MCTL_MSB_BANK		0x00	/* implemented */
#define ULTRA_MCTL_MSB_MODWHEEL         0x01	/* implemented */
#define ULTRA_MCTL_MSB_BREATH           0x02
#define ULTRA_MCTL_MSB_FOOT             0x04
#define ULTRA_MCTL_MSB_PORTNAMENTO_TIME 0x05
#define ULTRA_MCTL_MSB_DATA_ENTRY	0x06	/* implemented */
#define ULTRA_MCTL_MSB_MAIN_VOLUME      0x07	/* implemented */
#define ULTRA_MCTL_MSB_BALANCE          0x08
#define ULTRA_MCTL_MSB_PAN              0x0a	/* implemented */
#define ULTRA_MCTL_MSB_EXPRESSION       0x0b	/* implemented */
#define ULTRA_MCTL_MSB_EFFECT1		0x0c
#define ULTRA_MCTL_MSB_EFFECT2		0x0d
#define ULTRA_MCTL_MSB_GENERAL_PURPOSE1 0x10
#define ULTRA_MCTL_MSB_GENERAL_PURPOSE2 0x11
#define ULTRA_MCTL_MSB_GENERAL_PURPOSE3 0x12
#define ULTRA_MCTL_MSB_GENERAL_PURPOSE4 0x13
#define ULTRA_MCTL_LSB_BANK		0x20	/* implemented */
#define ULTRA_MCTL_LSB_MODWHEEL         0x21	/* ignored */
#define ULTRA_MCTL_LSB_BREATH           0x22
#define ULTRA_MCTL_LSB_FOOT             0x24
#define ULTRA_MCTL_LSB_PORTNAMENTO_TIME 0x25
#define ULTRA_MCTL_LSB_DATA_ENTRY	0x26	/* ignored */
#define ULTRA_MCTL_LSB_MAIN_VOLUME      0x27	/* ignored */
#define ULTRA_MCTL_LSB_BALANCE          0x28
#define ULTRA_MCTL_LSB_PAN              0x2a	/* ignored */
#define ULTRA_MCTL_LSB_EXPRESSION       0x2b	/* ignored */
#define ULTRA_MCTL_LSB_EFFECT1		0x2c
#define ULTRA_MCTL_LSB_EFFECT2		0x2d
#define ULTRA_MCTL_LSB_GENERAL_PURPOSE1 0x30
#define ULTRA_MCTL_LSB_GENERAL_PURPOSE2 0x31
#define ULTRA_MCTL_LSB_GENERAL_PURPOSE3 0x32
#define ULTRA_MCTL_LSB_GENERAL_PURPOSE4 0x33
#define ULTRA_MCTL_SUSTAIN              0x40	/* implemented */
#define ULTRA_MCTL_PORNAMENTO           0x41
#define ULTRA_MCTL_SOSTENUTO            0x42
#define ULTRA_MCTL_SOFT_PEDAL           0x43
#define ULTRA_MCTL_LEGATO_FOOTSWITCH	0x44
#define ULTRA_MCTL_HOLD2                0x45
#define ULTRA_MCTL_SC1_SOUND_VARIATION	0x46
#define ULTRA_MCTL_SC2_TIMBRE		0x47
#define ULTRA_MCTL_SC3_RELEASE_TIME	0x48
#define ULTRA_MCTL_SC4_ATTACK_TIME	0x49
#define ULTRA_MCTL_SC5_BRIGHTNESS	0x4a
#define ULTRA_MCTL_SC6			0x4b
#define ULTRA_MCTL_SC7			0x4c
#define ULTRA_MCTL_SC8			0x4d
#define ULTRA_MCTL_SC9			0x4e
#define ULTRA_MCTL_SC10			0x4f
#define ULTRA_MCTL_GENERAL_PURPOSE5     0x50
#define ULTRA_MCTL_GENERAL_PURPOSE6     0x51
#define ULTRA_MCTL_GENERAL_PURPOSE7     0x52
#define ULTRA_MCTL_GENERAL_PURPOSE8     0x53
#define ULTRA_MCTL_PORNAMENTO_CONTROL	0x54
#define ULTRA_MCTL_E1_REVERB_DEPTH	0x5b	/* implemented - InterWave only */
#define ULTRA_MCTL_E2_TREMOLO_DEPTH	0x5c
#define ULTRA_MCTL_E3_CHORUS_DEPTH	0x5d	/* implemented - InterWave only */
#define ULTRA_MCTL_E4_DETUNE_DEPTH	0x5e
#define ULTRA_MCTL_E5_PHASER_DEPTH	0x5f
#define ULTRA_MCTL_DATA_INCREMENT       0x60	/* implemented */
#define ULTRA_MCTL_DATA_DECREMENT       0x61	/* implemented */
#define ULTRA_MCTL_NONREG_PARM_NUM_LSB  0x62	/* implemented */
#define ULTRA_MCTL_NONREG_PARM_NUM_MSB  0x63	/* implemented */
#define ULTRA_MCTL_REGIST_PARM_NUM_LSB  0x64	/* implemented */
#define ULTRA_MCTL_REGIST_PARM_NUM_MSB  0x65	/* implemented */
#define ULTRA_MCTL_ALL_SOUNDS_OFF	0x78	/* implemented */
#define ULTRA_MCTL_RESET_CONTROLLERS	0x79	/* implemented */
#define ULTRA_MCTL_LOCAL_CONTROL_SWITCH	0x7a
#define ULTRA_MCTL_ALL_NOTES_OFF	0x7b	/* implemented */
#define ULTRA_MCTL_OMNI_OFF		0x7c	/* same as all notes off */
#define ULTRA_MCTL_OMNI_ON		0x7d	/* same as all notes off */
#define ULTRA_MCTL_MONO1		0x7e	/* same as all notes off */
#define ULTRA_MCTL_MONO2		0x7f	/* same as all notes off */
/* extended controllers (only for ULTRA driver) */
#define ULTRA_MCTL_TREMOLO		0x80
#define ULTRA_MCTL_VIBRATO		0x81
#define ULTRA_MCTL_LEGATO		0x82

/*
 *  Various structures
 */

typedef struct ULTRA_PCM_INFO ultra_pcm_info_t;

/****************************************************************************
 *                                                                          *
 *      Section for driver control interface - /dev/ultracontrol?           *
 *                                                                          *
 ****************************************************************************/

#define ULTRA_CTL_VERSION		ULTRA_PROTOCOL_VERSION( 1, 0, 0 )

#define ULTRA_CTL_GCAPS_MIDI		0x0000001	/* driver have MIDI interface */

#define ULTRA_CTL_LCAPS_SYNTH		0x0000001	/* soundcard have synthesizer */
#define ULTRA_CTL_LCAPS_PCM0		0x0000010	/* soundcard have PCM - device 0 */
#define ULTRA_CTL_LCAPS_PCM1		0x0000020	/* soundcard have PCM - device 1 */
#define ULTRA_CTL_LCAPS_MIXER0		0x0001000	/* soundcard have MIXER - device 0 */
#define ULTRA_CTL_LCAPS_MIXER1		0x0002000	/* soundcard have MIXER - device 1 */
#define ULTRA_CTL_LCAPS_RAWMIDI0	0x0100000	/* soundcard have RAW MIDI - device 0 */
#define ULTRA_CTL_LCAPS_RAWMIDI1	0x0200000	/* soundcard have RAW MIDI - device 1 */
#define ULTRA_CTL_LCAPS_RAWFM		0x1000000	/* soundcard have RAW FM/OPL3 */

#define ULTRA_CTL_TYPE_INTEGER		0x0000000
#define ULTRA_CTL_TYPE_STRING		0x0000001
#define ULTRA_CTL_TYPE_BOOLEAN		0x0000002

struct ULTRA_CTL_HW_INFO {
  unsigned int type;		/* type of card - look to ULTRA_CTL_TYPE_XXXX */
  unsigned int gcaps;		/* look to ULTRA_CTL_GCAPS_XXXX */
  unsigned int lcaps;		/* look to ULTRA_CTL_LCAPS_XXXX */
  char id[8];			/* ID of card (user selectable) */
  char info1[80];		/* info text 1 */
  char info2[80]; 		/* info text 2 */
  unsigned char reserved[128];	/* reserved for future */
};

struct ULTRA_CTL_HW_SETUP {
  unsigned char param[16];	/* parameter name */
  union {
    unsigned int integer;	/* integer value */
    unsigned char string[128];	/* string value */
  } data;
};

struct ULTRA_CTL_HW_SETUP_LIST {
  unsigned int params_count;	/* total count of parameters */
  unsigned int param_index;	/* current index of parameter (must be filled in user space) */
  unsigned char param[16];	/* returned name of parameter */
  unsigned int param_type;	/* returned type of parameter (see to ULTRA_CTL_TYPE_XXXX) */
  union {
    struct {
      unsigned int min;		/* minimum value */
      unsigned int max; 	/* maximum value */
      unsigned int step;	/* step value (1-X) */
    } integer;
    struct {
      unsigned int maxlen;	/* max. length in chars (with '\0') */
    } string;
  } description;
  unsigned char reserved[16];	/* reserved for future use */
};

#define ULTRA_IOCTL_CTL_VERSION		_IOR ( 'U', 0x00, int )
#define ULTRA_IOCTL_CTL_HW_INFO		_IOR ( 'U', 0x01, struct ULTRA_CTL_HW_INFO )
#define ULTRA_IOCTL_CTL_HW_SETUP	_IOW ( 'U', 0x02, struct ULTRA_CTL_HW_SETUP )
#define ULTRA_IOCTL_CTL_HW_GET_SETUP	_IOWR( 'U', 0x02, struct ULTRA_CTL_HW_SETUP )
#define ULTRA_IOCTL_CTL_HW_SETUP_LIST	_IOWR( 'U', 0x03, struct ULTRA_CTL_HW_SETUP_LIST )
#if 0
#define ULTRA_IOCTL_CTL_SYNTH_INFO	_IOR ( 'U', 0x10, ultra_synth_info_t )
#endif
#define ULTRA_IOCTL_CTL_PCM_INFO	_IOWR( 'U', 0x11, ultra_pcm_info_t )

/****************************************************************************
 *                                                                          *
 *                MIXER interface - /dev/ultramixer?                        *
 *                                                                          *
 ****************************************************************************/
 
#define ULTRA_MIXER_VERSION		ULTRA_PROTOCOL_VERSION( 1, 0, 0 )

					/* max 12 chars (with '\0') */
#define ULTRA_MIXER_ID_MASTER		"Master"
#define ULTRA_MIXER_ID_SYNTHESIZER	"Synth"
#define ULTRA_MIXER_ID_SYNTHESIZER1	"Synth 1"
#define ULTRA_MIXER_ID_EFFECT		"Effect"
#define ULTRA_MIXER_ID_PCM		"PCM"
#define ULTRA_MIXER_ID_PCM1		"PCM 1"
#define ULTRA_MIXER_ID_LINE		"Line-In"
#define ULTRA_MIXER_ID_MIC		"MIC"
#define ULTRA_MIXER_ID_CD		"CD"
#define ULTRA_MIXER_ID_GAIN		"Record-Gain"
#define ULTRA_MIXER_ID_IGAIN		"In-Gain"
#define ULTRA_MIXER_ID_OGAIN		"Out-Gain"
#define ULTRA_MIXER_ID_LOOPBACK		"Loopback"
#define ULTRA_MIXER_ID_SPEAKER		"PC Speaker"
#define ULTRA_MIXER_ID_AUXA		"Aux A"
#define ULTRA_MIXER_ID_AUXB		"Aux B"
#define ULTRA_MIXER_ID_AUXC		"Aux C"

#define ULTRA_MIXER_INFO_CAP_EXCL_RECORD 0x00000001

#define ULTRA_MIXER_DINFO_CAP_RECORD	0x00000001
#define ULTRA_MIXER_DINFO_CAP_STEREO	0x00000002
#define ULTRA_MIXER_DINFO_CAP_MUTE	0x00000004

#define ULTRA_MIXER_FLG_RECORD		0x00000001
#define ULTRA_MIXER_FLG_MUTE_LEFT	0x00010000
#define ULTRA_MIXER_FLG_MUTE_RIGHT	0x00020000
#define ULTRA_MIXER_FLG_MUTE		0x00030000

#define ULTRA_MIXER_S_NONE				0
#define ULTRA_MIXER_S_IW				1

#define ULTRA_MIXER_S_IW_SERIAL_NONE			0
#define ULTRA_MIXER_S_IW_SERIAL_DSP_TO_RECORD		1
#define ULTRA_MIXER_S_IW_SERIAL_DSP_TO_PLAYBACK		2
#define ULTRA_MIXER_S_IW_SERIAL_RECORD_TO_PLAYBACK	3
#define ULTRA_MIXER_S_IW_SERIAL_DSP_TO_EXTOUT		4
#define ULTRA_MIXER_S_IW_SERIAL_RECORD_TO_EXTOUT	5
#define ULTRA_MIXER_S_IW_SERIAL_EXTIN_TO_PLAYBACK_1	6
#define ULTRA_MIXER_S_IW_SERIAL_EXTIN_TO_PLAYBACK_2	7

struct ULTRA_MIXER_INFO {
  unsigned int version;		/* version of soundcard */
  unsigned int channels;	/* count of mixer devices */
  unsigned int caps;		/* some flags about this device (ULTRA_MIXER_INFO_XXXX) */
  unsigned char id[32];		/* ID of this mixer */
  unsigned char info[64];	/* some info text about mixer */
  char reserved[ 32 ];		/* reserved for future use */
};

struct ULTRA_MIXER_CHANNEL_INFO {
  unsigned int channel;		/* device # (filled by application) */
  unsigned char name[12];	/* name of this device */
  unsigned int caps;		/* some flags about this device (ULTRA_MIXER_DINFO_XXXX) */
};

struct ULTRA_MIXER_CHANNEL {
  unsigned int channel;		/* channel # (filled by application) */
  unsigned int flags;		/* some flags to read/write (ULTRA_MIXER_FLG_XXXX) */
  unsigned char left;		/* 0 - 100 */
  unsigned char right;		/* 0 - 100 */
};

struct ULTRA_MIXER_SPECIAL {
  unsigned int what;
  union {
    unsigned char bytes[32];
    unsigned short words[16];
    unsigned int dwords[8];
    struct {
      unsigned char serial;
    } interwave;
  } data;
  unsigned char reserved[32];
};

#define ULTRA_MIXER_IOCTL_VERSION	_IOR ( 'U', 0, int )
#define ULTRA_MIXER_IOCTL_CHANNELS	_IOR ( 'U', 0, int )
#define ULTRA_MIXER_IOCTL_INFO		_IOR ( 'U', 1, struct ULTRA_MIXER_INFO )
#define ULTRA_MIXER_IOCTL_CHANNEL_INFO	_IOWR( 'U', 2, struct ULTRA_MIXER_CHANNEL_INFO )
#define ULTRA_MIXER_IOCTL_CHANNEL_READ	_IOR ( 'U', 3, struct ULTRA_MIXER_CHANNEL )
#define ULTRA_MIXER_IOCTL_CHANNEL_WRITE	_IOW ( 'U', 3, struct ULTRA_MIXER_CHANNEL )
#define ULTRA_MIXER_IOCTL_SPECIAL_READ	_IOR ( 'U', 4, struct ULTRA_MIXER_SPECIAL )
#define ULTRA_MIXER_IOCTL_SPECIAL_WRITE	_IOW ( 'U', 4, struct ULTRA_MIXER_SPECIAL )

/*****************************************************************************
 *                                                                           *
 *           Digital Audio (PCM) interface - /dev/ultrapcm?                  *
 *                                                                           *
 *****************************************************************************/

#define ULTRA_PCM_VERSION		ULTRA_PROTOCOL_VERSION( 1, 0, 0 )

#define ULTRA_PCM_FMT_QUERY		0x00000000
#define ULTRA_PCM_FMT_MU_LAW		0x00000001
#define ULTRA_PCM_FMT_A_LAW		0x00000002
#define ULTRA_PCM_FMT_IMA_ADPCM		0x00000004
#define ULTRA_PCM_FMT_U8		0x00000008
#define ULTRA_PCM_FMT_S16_LE		0x00000010
#define ULTRA_PCM_FMT_S16_BE		0x00000020
#define ULTRA_PCM_FMT_S8		0x00000040
#define ULTRA_PCM_FMT_U16_LE		0x00000080
#define ULTRA_PCM_FMT_U16_BE		0x00000100
#define ULTRA_PCM_FMT_MPEG		0x00000200

#define ULTRA_PCM_INFO_CODEC		0x00000001
#define ULTRA_PCM_INFO_DSP		ULTRA_PCM_INFO_CODEC
#define ULTRA_PCM_INFO_MMAP		0x00000002
#define ULTRA_PCM_INFO_PLAYBACK		0x00000100
#define ULTRA_PCM_INFO_RECORD		0x00000200
#define ULTRA_PCM_INFO_DUPLEX		0x00000400
#define ULTRA_PCM_INFO_DUPLEX_LIMIT	0x00000800
#define ULTRA_PCM_INFO_PLAYBACK_BATCH	0x00001000
#define ULTRA_PCM_INFO_RECORD_BATCH	0x00002000

#define ULTRA_PCM_MASK_PLAYBACK		0x0001
#define ULTRA_PCM_MASK_RECORD		0x0002
#define ULTRA_PCM_MASK_DUPLEX		(ULTRA_PCM_MASK_PLAYBACK|ULTRA_PCM_MASK_RECORD)
#define ULTRA_PCM_MASK_BOTH		ULTRA_PCM_MASK_DUPLEX

#define ULTRA_PCM_ENABLE_RECORD		0x00000001
#define ULTRA_PCM_ENABLE_PLAYBACK	0x00000002

#define ULTRA_PCM_CAP_REVISION		0x000000ff
#define ULTRA_PCM_CAP_DUPLEX		0x00000100
#define ULTRA_PCM_CAP_REALTIME		0x00000200
#define ULTRA_PCM_CAP_BATCH		0x00000400
#define ULTRA_PCM_CAP_COPROC		0x00000800
#define ULTRA_PCM_CAP_TRIGGER		0x00001000
#define ULTRA_PCM_CAP_MMAP		0x00002000

struct ULTRA_PCM_INFO {
  unsigned int device;			/* device # */
  unsigned int flags;			/* see to ULTRA_PCM_INFO_XXXX */
  unsigned int formats_play;		/* supported formats */
  unsigned int formats_record;		/* supported formats */
  unsigned int max_rate_play;		/* maximal rate (in Hz) */
  unsigned int max_rate_record;		/* maximal rate (in Hz) */
  unsigned int max_channels_play;	/* maximal channels # */ 
  unsigned int max_channels_record;	/* maximal channels # */
  unsigned int dma_size_play;		/* playback DMA size */
  unsigned int dma_size_record;		/* record DMA size */
  unsigned char name[80];		/* name of this device */
  unsigned int reserved[64];		/* reserved for future... */
};

struct ULTRA_PCM_BUFFER_INFO {
  int fragments;	/* # of available fragments (partially used ones not counted) */
  int fragstotal;	/* Total # of fragments allocated */
  int fragsize;		/* Size of a fragment in bytes */
  int bytes;		/* Available space in bytes (includes partially used fragments) */ 
};

struct ULTRA_PCM_COUNT_INFO {
  int bytes;		/* Total # of bytes processed */
  int blocks;		/* # of fragment transitions since last time */
  int ptr;		/* Current DMA pointer value */
};

struct ULTRA_PCM_BUFFER_DESCRIPTION {
  unsigned char *buffer;
  int size;
};

/* OSS compatible part */
#define ULTRA_PCM_IOCTL_RESET		_IO  ( 'P', 0 )
#define ULTRA_PCM_IOCTL_SYNC		_IO  ( 'P', 1 )
#define ULTRA_PCM_IOCTL_RATE		_IOWR( 'P', 2, int )
#define ULTRA_PCM_IOCTL_GETRATE		_IOR ( 'P', 2, int )
#define ULTRA_PCM_IOCTL_STEREO		_IOWR( 'P', 3, int )
#define ULTRA_PCM_IOCTL_GETBLKSIZE	_IOWR( 'P', 4, int )
#define ULTRA_PCM_IOCTL_FORMAT		_IOWR( 'P', 5, int )
#define ULTRA_PCM_IOCTL_GETFORMAT	_IOR ( 'P', 5, int )
#define ULTRA_PCM_IOCTL_CHANNELS	_IOWR( 'P', 6, int )
#define ULTRA_PCM_IOCTL_GETCHANNELS	_IOR ( 'P', 6, int )
#define ULTRA_PCM_IOCTL_FILTER		_IOWR( 'P', 7, int )
#define ULTRA_PCM_IOCTL_GETFILTER	_IOR ( 'P', 7, int )
#define ULTRA_PCM_IOCTL_POST		_IO  ( 'P', 8 )
#define ULTRA_PCM_IOCTL_SUBDIVIDE	_IOWR( 'P', 9, int )
#define ULTRA_PCM_IOCTL_SETFRAGMENT	_IOWR( 'P', 10, int )
#define ULTRA_PCM_IOCTL_GETFORMATS	_IOR ( 'P', 11, int )
#define ULTRA_PCM_IOCTL_GETPBKSPACE	_IOR ( 'P', 12, struct ULTRA_PCM_BUFFER_INFO )
#define ULTRA_PCM_IOCTL_GETRECSPACE	_IOR ( 'P', 13, struct ULTRA_PCM_BUFFER_INFO )
#define ULTRA_PCM_IOCTL_NONBLOCK	_IO  ( 'P', 14 )
#define ULTRA_PCM_IOCTL_GETCAPS		_IOR ( 'P', 15, int )
#define ULTRA_PCM_IOCTL_GETTRIGGER	_IOR ( 'P', 16, int )
#define ULTRA_PCM_IOCTL_SETTRIGGER	_IOW ( 'P', 16, int )
#define ULTRA_PCM_IOCTL_GETRECPTR	_IOR ( 'P', 17, struct ULTRA_PCM_COUNT_INFO )
#define ULTRA_PCM_IOCTL_GETPBKPTR	_IOR ( 'P', 18, struct ULTRA_PCM_COUNT_INFO )
#define ULTRA_PCM_IOCTL_MAPRECBUFFER	_IOR ( 'P', 19, struct ULTRA_PCM_BUFFER_DESCRIPTION )
#define ULTRA_PCM_IOCTL_MAPPBKBUFFER	_IOR ( 'P', 20, struct ULTRA_PCM_BUFFER_DESCRIPTION )
#define ULTRA_PCM_IOCTL_SYNCRO		_IO  ( 'P', 21 )
#define ULTRA_PCM_IOCTL_DUPLEX		_IO  ( 'P', 22 )
/* ULTRA native ioctl's */
#define ULTRA_PCM_IOCTL_VERSION		_IOR ( 'P', 240, int )
#define ULTRA_PCM_IOCTL_INFO		_IOR ( 'P', 241, struct ULTRA_PCM_INFO )
#define ULTRA_PCM_IOCTL_MASK		_IOW ( 'P', 241, int )

/*
 *
 */
 
#endif /* __ULTRA_H */
